import sqlite3
from _thread import get_ident
import xbmc
import xbmcgui
from helper import utils
from . import emby_db, video_db, music_db, texture_db, common_db

DBConnectionsRW = {}
DBConnectionsRO = {}

def set_walmode():
    for DBID, DBFile in list(utils.DatabaseFiles.items()):
        xbmc.log(f"EMBY.database.dbio: ---> WAL mode: {DBID}", 0) # LOGDEBUG

        if 'version' in DBID or DBFile.decode('utf-8').find('emby') != -1:
            continue

        if DBID in DBConnectionsRW:
            while DBConnectionsRW[DBID][1]:  #Wait for db unlock
                xbmc.log(f"EMBY.database.dbio: DBOpenRW: Waiting WAL mode {DBID}", 1) # LOGINFO
                utils.sleep(1)
        else:
            globals()["DBConnectionsRW"][DBID] = [None, False]

        globals()["DBConnectionsRW"][DBID] = [sqlite3.connect(DBFile, timeout=999999), True]

        try:
            DBConnectionsRW[DBID][0].execute("PRAGMA journal_mode=WAL")
        except Exception as Error:
            xbmc.log(f"EMBY.database.dbio: Cannot det WAL mode: {DBID} / {Error}", 3) # LOGERROR

        DBConnectionsRW[DBID][0].close()
        globals()["DBConnectionsRW"][DBID][1] = False
        xbmc.log(f"EMBY.database.dbio: ---< WAL mode: {DBID}", 0) # LOGDEBUG

def DBVacuum():
    xbmc.executebuiltin('Dialog.Close(addoninformation)')
    ProgressBar = xbmcgui.DialogProgressBG()
    ProgressBar.create(utils.Translate(33199), utils.Translate(33436))
    TotalItems = len(utils.DatabaseFiles) / 100
    Index = 1

    for DBID, DBFile in list(utils.DatabaseFiles.items()):
        ProgressBar.update(int(Index / TotalItems), utils.Translate(33436), str(DBID))

        if 'version' in DBID:
            continue

        xbmc.log(f"EMBY.database.dbio: ---> DBVacuum: {DBID}", 0) # LOGDEBUG

        if DBID in DBConnectionsRW:
            while DBConnectionsRW[DBID][1]:  #Wait for db unlock
                xbmc.log(f"EMBY.database.dbio: DBOpenRW: Waiting Vacuum {DBID}", 1) # LOGINFO
                utils.sleep(1)
        else:
            globals()["DBConnectionsRW"][DBID] = [None, False]

        globals()["DBConnectionsRW"][DBID] = [sqlite3.connect(DBFile, timeout=999999), True]

        if DBID == "music":
            curser = DBConnectionsRW[DBID][0].cursor()
            curser.execute("DELETE FROM removed_link")
            curser.close()
            DBConnectionsRW[DBID][0].commit()
            DBConnectionsRW[DBID][0].close()
            globals()["DBConnectionsRW"][DBID][0] = sqlite3.connect(DBFile, timeout=999999)

        DBConnectionsRW[DBID][0].execute("VACUUM")
        DBConnectionsRW[DBID][0].cursor().close()
        DBConnectionsRW[DBID][0].close()
        globals()["DBConnectionsRW"][DBID][1] = False
        xbmc.log(f"EMBY.database.dbio: ---< DBVacuum: {DBID}", 0) # LOGDEBUG
        Index += 1

    ProgressBar.close()
    del ProgressBar

def DBOpenRO(DBID, TaskId):
    DBIDThreadID = f"{DBID}{TaskId}{get_ident()}"

    try:
        globals()["DBConnectionsRO"][DBIDThreadID] = sqlite3.connect(f"file:{utils.DatabaseFiles[DBID].decode('utf-8')}?immutable=1&mode=ro", uri=True, timeout=999999) #, check_same_thread=False
    except Exception as Error:
        xbmc.log(f"EMBY.database.dbio: Database IO: {DBID} / {TaskId} / {Error}", 3) # LOGERROR
        return None

    xbmc.log(f"EMBY.database.dbio: ---> DBRO: {DBIDThreadID}", 0) # LOGDEBUG

    if DBID == 'video':
        return video_db.VideoDatabase(DBConnectionsRO[DBIDThreadID].cursor())

    if DBID == 'music':
        return music_db.MusicDatabase(DBConnectionsRO[DBIDThreadID].cursor())

    if DBID == 'texture':
        return texture_db.TextureDatabase(DBConnectionsRO[DBIDThreadID].cursor())

    if DBID in ('epg', 'tv'):
        return common_db.CommonDatabase(DBConnectionsRO[DBIDThreadID].cursor())

    return emby_db.EmbyDatabase(DBConnectionsRO[DBIDThreadID].cursor())

def DBCloseRO(DBID, TaskId):
    DBIDThreadID = f"{DBID}{TaskId}{get_ident()}"

    if DBIDThreadID in DBConnectionsRO:
        DBConnectionsRO[DBIDThreadID].cursor().close()
        DBConnectionsRO[DBIDThreadID].close()
        xbmc.log(f"EMBY.database.dbio: ---< DBRO: {DBIDThreadID}", 0) # LOGDEBUG
    else:
        xbmc.log(f"EMBY.database.dbio: DBIDThreadID not found {DBIDThreadID}", 3) # LOGERROR

def DBOpenRW(Databases, TaskId, SQLs):
    DBIDs = Databases.split(",")

    for DBID in DBIDs:
        if DBID == "none":
            continue

        if DBID in DBConnectionsRW:
            while DBConnectionsRW[DBID][1]:  #Wait for db unlock
                xbmc.log(f"EMBY.database.dbio: DBOpenRW: Waiting {DBID} / {TaskId}", 1) # LOGINFO
                utils.sleep(1)
        else:
            globals()["DBConnectionsRW"][DBID] = [None, False]

        globals()["DBConnectionsRW"][DBID] = [sqlite3.connect(utils.DatabaseFiles[DBID].decode('utf-8'), timeout=999999), True]
        xbmc.log(f"EMBY.database.dbio: ---> DBRW: {DBID}/{TaskId}", 0) # LOGDEBUG

        if DBID == 'video':
            SQLs[DBID] = video_db.VideoDatabase(DBConnectionsRW[DBID][0].cursor())
        elif DBID == 'music':
            SQLs[DBID] = music_db.MusicDatabase(DBConnectionsRW[DBID][0].cursor())
        elif DBID == 'texture':
            SQLs[DBID] = texture_db.TextureDatabase(DBConnectionsRW[DBID][0].cursor())
        elif DBID == 'epg':
            SQLs[DBID] = common_db.CommonDatabase(DBConnectionsRW[DBID][0].cursor())
        elif DBID == 'tv':
            SQLs[DBID] = common_db.CommonDatabase(DBConnectionsRW[DBID][0].cursor())
        else:
            SQLs["emby"] = emby_db.EmbyDatabase(DBConnectionsRW[DBID][0].cursor())

    return SQLs

def DBCloseRW(Databases, TaskId, SQLs):
    DBIDs = Databases.split(",")

    for DBID in DBIDs:
        if DBID == "none":
            continue

        changes = DBConnectionsRW[DBID][0].total_changes

        if changes:
            DBConnectionsRW[DBID][0].commit()

        DBConnectionsRW[DBID][0].cursor().close()
        DBConnectionsRW[DBID][0].close()
        globals()["DBConnectionsRW"][DBID][1] = False

        if DBID in ('video', 'music', 'texture', 'epg', 'tv'):
            SQLs[DBID] = None
        else:
            SQLs["emby"] = None

        xbmc.log(f"EMBY.database.dbio: ---< DBRW: {DBID} / {changes} / {TaskId} rows updated on db close", 0) # LOGDEBUG

    return SQLs
